unit UnitClipQueue;
{
    Purpose:
        Encapsulate all the rules of of the queue
        - number of items allowed
        - deleting icons on item removal (no memory leaks)
        - etc
        Handle a separate queue of items removed from the ClipQueue

    Updates:
        GetClipboardFormatName - check for failure
        -------------------
        Fix for 0 sized TPagedStringQueue
        
        ------------------
        Better way to enumerate the clipboard formats - single API call
        Workaround for WinXP when retreiving from clipboard


        ------------------
        Ability to save/load a clipboard item (TClipItem)
        Full color ICons

        ------------------
        Fix for DragQuery API (ClipItem) on Win9x machines.

        Revamped naming a little - removed some TClipItem wrapper functions
        from the queues (that weren't really a queue function).
        New Objects to simplify support for Rich Text, Unicode, and
        more. TClipData & TClipItem


        Added support for storing many data items along with text version of
        and item.
        Added support for loading a StringQueue (for Remove Items)

        --------------
        Created a basic StringQueue that's just a FIFO to be
        used for a RemovedQueue object.
        MoveDuplicateTop logic moved here (where it should be)
        Ignore duplicates durring Add or Instert
}


{ TODO: Investigate large HTML -> RichText clips failing if large}
{///////////////}
{//}interface{//}
{///////////////}
uses unitmisc, classes, Windows, Graphics {TPicture}, inifiles {hash}, UnitTWideChar;

var ClipDataDefaultIcon : HICON;
{
    TClipData
        - data associated with a clipboard item ( aka TClipItem )
}
type TClipData = class(TObject)
    private
        size : cardinal;
        s : string;
        h : HICON;
        timestamp : TDateTime;

    public
        constructor Create(s : string; h : HICON);
        destructor Destroy; override;
        function GetString : string;
        function GetHICON : HICON;
        function GetHICONAbsolute : HICON; {used so the default icon is not saved}
        procedure SetString(s : string);
        procedure SetHICON(h : HICON);
        function GetCreationDate : TDateTime;
end;
{
    TClipItem
        - Get's data from the clipboard in native form,
        put's data on the clipboard, Keeps a plain text version of the item,
        and associates an icon with the handle (see TClipData)
}
const CF_FILE_RICHTEXT = CF_PRIVATEFIRST;
const CF_FILE_HTML = CF_PRIVATEFIRST + 1;

type TClipItem = class(TObject)
    private

        CFormat : WORD; {The format & Handle of the copied clipboard item}
        CHandle : THandle;  {The bare essentials for a clipboard item}

        function GetFilenamesAsText(h : THandle) : string;
        procedure CleanupMemory;
        //procedure PlaceOnClipboard;
    public

        CData : TClipdata; {Data relating to the clipboard item}
                           {Used to store a text version of the file and
                            save the ICON of the program that this text came from}
        constructor Create;
        destructor Destroy; override;


        function GetClipboardItem(hi : HICON; OverrideFormat : WORD = 0; SizeLimit : cardinal = $FFFF) : cardinal; overload;
        procedure SetFromWideChar(wc : TWideChar);

        function GetAsText : string;
        function HasText : boolean;
        procedure OverrideTextVersionOfItem(s : string);
        function GetHandle : THandle;
        function GetDataSize : cardinal;
        function GetFormat : cardinal;
        function GetFormatName(AccessHandle : boolean = true) : string;
        procedure GetDIB(pic : TPicture);
        procedure GetRichText(var s : string);

        procedure SaveToFile(path : string; index : integer);
        procedure LoadFromFIle(path : string; index : integer);
        function GetFilename(path : string; index : integer) : string;

        procedure SaveIconToFile(path : string; index : integer; sufix : string = '');
        procedure LoadIconFromFile(path : string; index : integer; sufix : string = '');
        function GeticonFilename(path : string; index : integer; sufix : string = '') : string;
        {util}
        //function DupHandle(h : Thandle; var sizeh : cardinal; SizeLimit : boolean = false) : Thandle;
end;

{
    TStringQueue
        - Basic FIFO for strings
        - Enforces a total size
        - Override for populating the list
    NOTES:
        TStringList container because I don't want to deal with overriding
        all the many ways the list can be altered. Lazy, yes I am, but it also
        saves me from myself.
}
type TStringQueue = class(TObject)
    protected
        sl : TStringList;
        qSize : cardinal;
    public
        constructor Create;
        destructor Destroy; override;
        // init the rule
        procedure SetQueueSize(size : longint);
        function GetQueueSize : longint;
        // Add - for loading history, no size check
        // InsertAtStart - for FrmClipboardManagaer
        procedure AddNoSizeCheck(s : string);
        procedure InsertAtStart(s : string); overload;

        function GetQueueCount : cardinal;
        function GetItemText(index : cardinal) : string;


        procedure DeleteItem(index : cardinal); virtual;
        procedure ClearQueue; virtual;
end;

{
    TClipQueue
        - A string Queue that ass
}
type TClipQueue = class(TStringQueue)
    private
        MoveDuplicateTop : boolean;


    public
        constructor Create;
        destructor Destroy; override;


        {config options}
        procedure SetMoveDuplicateTop(enable: boolean);

        procedure MoveToStart(index : cardinal);
        procedure Move(oldIndex : cardinal; newIndex : cardinal);

        {add & retreive operations}
        procedure GetQueueItems(items : TStrings);
        procedure SetQueueItems(items : TStrings);

        procedure AddNoSizeCheck(s : string; ci : TClipItem = nil); virtual;
        procedure InsertAtStart(ci : TClipItem; casesensative : boolean = false); overload;
        function GetClipItem(index : cardinal) : TClipItem;


        {find & delete}
        function IndexOf(s : string) : integer;  overload;
        function IndexOf(ci : TClipItem) : integer; overload;
        procedure DeleteItem(index : cardinal); override;
        procedure ClearQueue; override;
end;


//
// buffer between the ClipQueue and the PagedStringQueue
//
type TTempClipQueue = class(TClipQueue)
      procedure DeleteItem(index : cardinal); override;
      procedure DeleteItem2(index : cardinal; NoFree : boolean = false); overload;
      procedure InsertAtStart(ci : TClipItem); overload;
      procedure AddNoSizeCheck(s : string; ci : TClipItem = nil); override;
end;


procedure ResyncTempClipQueue;



{
    TPagedStringQueue
        Poor Engineering 101 here, but oh well. (Does not inherit
        from TStringQueue)

        - Circular queue of items 0-N but the Queue "start" is the first
        item in the list
        - Item 0 is the oldest item, Item (GetQueueCount - 1) is the newest
        item
}
const PAGED_EXT = '.ac';
const PAGED_CLIP_EXT = '.acz';
const PAGED_STATEFILE = 'start' + PAGED_EXT;
const PAGED_ICON_EXT = '.bmp';

type TPagedStringQueue = class(TObject)
    private
        base : string;
        filename : string;
        cache : string;
        iconcache : string;
        qSize : cardinal;
        qStart : cardinal;
        qCount : cardinal;

        ci : TClipITem;

        iconcachecrc : THashedStringList;

        procedure SaveItem(s : string; index: cardinal; ci : TClipItem = nil);

        function IsEmptyItem(index : cardinal) : boolean;
        function IsEmptyItemAbsolute(absoluteIndex : cardinal) : boolean;
        function IsEmptyItemClipAbsolute(absoluteIndex : cardinal) : boolean;

        procedure RemoveOldestItem;
        procedure SaveQueueState;
        //procedure ReIndexQueue;

        function IndexTranslate(index : cardinal) : cardinal;


        function GetItemAbsolute(absoluteIndex : cardinal) : string;
        function GetItemClipAbsolute(absoluteIndex : cardinal) : TClipItem;

        //function GetFilenameOld(index : cardinal) : string;
        //function GetFilenameNew(index : cardinal) : string;
        function GetFilenameAbsolute(absoluteindex : cardinal) : string;
        function GetFilenameAbsoluteOld(absoluteIndex : cardinal) : string;
        function GetFilenameClip(index : cardinal) : string;
        function GetFilenameClipAbsolute(absoluteIndex : cardinal) : string;
        function GetFilenameIcon1Absolute(absoluteIndex : cardinal) : string;
        function GetFilenameIcon2Absolute(absoluteIndex : cardinal) : string;
    public

        constructor Create(filename : string; folder : string = '');
        destructor Destroy; override;


        procedure InsertAtStart(s : string; ci : TClipItem = nil); overload;
        function GetQueueCount : cardinal;
        function GetItemText(index : cardinal) : string;
        function GetItemClip(Index: cardinal): TClipItem;
        function GetItemDate(index : cardinal) : TDateTime;

        procedure ClearQueue;
        procedure SetQueueSize(size : cardinal);
end;



var ClipQueue : TClipQueue;
var RemovedQueue : TPagedStringQueue;
var TempClipQueue : TTempClipQueue;
{////////////////////}
{//}implementation{//}
{////////////////////}


uses UnitFrmMainPopup, Forms {For Application object},
    Clipbrd, StrUtils, SysUtils, ShellAPI, Dialogs, UnitPaste, Math, UnitFrmClipboardManager,
  UnitFrmConfig, UnitCRC32;

const FILE_EXT_BINARY = PAGED_CLIP_EXT;
const FILE_EXT = '.acy';

//-------------------
// (Con/De)structors
//-------------------
constructor TStringQueue.Create;
begin
    sl := TStringList.Create;
end;

destructor TStringQueue.Destroy;
begin
    MyFree(sl);
    inherited Destroy;
end;

procedure TStringQueue.SetQueueSize(size : longint);
begin
    qSize := size;
    while (sl.count > size) and (sl.count > 0) do
        self.DeleteItem(sl.count - 1);
end;

function TStringQueue.GetQueueSize: longint;
begin
    result := qSize;
end;



procedure TStringQueue.DeleteItem(index : cardinal);
begin
    sl.Delete(index);
end;

procedure TStringQueue.InsertAtStart(s : string);
begin
    sl.Insert(0,s);
    self.SetQueueSize(self.qSize);
end;

procedure TStringQueue.AddNoSizeCheck(s: string);
var i : longint;
begin
    i := sl.IndexOf(s);
    if (i = -1) then begin
        sl.Add(s);
    end;
end;




function TStringQueue.GetQueueCount : cardinal;
begin
    result := sl.count;
end;

function TStringQueue.GetItemText(index : cardinal) : string;
begin
    result := sl[index];
end;

procedure TStringQueue.ClearQueue;
var i : integer;
begin
    for i := (sl.count - 1) downto 0 do begin
        self.DeleteItem(i);
    end;
end;



//////////////////////////
//
// TClipQueue
//
//////////////////////////

//-------------------
// (Con/De)structors
//-------------------
constructor TClipQueue.Create;
begin
    sl := TStringList.Create;
end;

destructor TClipQueue.Destroy;
begin
    MyFree(sl);
    inherited Destroy;
end;


//-------------------
// Configuration
//-------------------
procedure TClipQueue.SetMoveDuplicateTop(enable: boolean);
begin
    self.MoveDuplicateTop := enable;
end;


//-------------------
// Public Interface
//-------------------


//
// items added to list
//
procedure TClipQueue.AddNoSizeCheck(s : string; ci : TClipItem = nil);
var i : longint;
begin
    // No dups, add move to top if configured to do so

    i := sl.IndexOf(s);
    if (i = -1) then begin
        sl.AddObject(s, ci);
    end else if (self.MoveDuplicateTop) then begin
        self.MoveToStart(i);
    end;
end;

procedure TClipQueue.InsertAtStart(ci : TClipItem; casesensative : boolean = false);
var i : integer;
    ci2 : TClipItem;
    cmp :integer;
begin
    sl.CaseSensitive :=  casesensative;
    i := sl.IndexOf(ci.GetAsText);
    if (i = -1) then begin
        sl.InsertObject(0, ci.GetAsText, ci);
        self.SetQueueSize(self.qSize);
    end else begin
        if (self.MoveDuplicateTop) then begin
            // delete an add it in its place if only differs in Case CASE case
            ci2 := self.GetClipItem(i);

            // use the "new" item and delete the old when case has changed
            // just remove the new item when identical
            cmp := CompareStr(ci2.GetAsText, ci.GetAsText);
            if  cmp <> 0 then begin
                DeleteItem(i);
                sl.InsertObject(0, ci.GetAsText, ci);
                self.SetQueueSize(self.qSize);
            end else begin
                self.MoveToStart(i);
                UnitMisc.MyDestroyIcon(ci.CData.GetHICONAbsolute);
                UnitMisc.MyFree(ci);
            end;
        end;
    end

end;



procedure TClipQueue.SetQueueItems(items : TStrings);
begin
    //
    // trim queue if needed

    self.ClearQueue;
    sl.AddStrings(items);
    self.SetQueueSize(self.qSize);
end;



function TClipQueue.IndexOf(s : string) : integer;
begin
    result := sl.IndexOf(s);
end;

function TClipQueue.IndexOf(ci: TClipItem): integer;
begin
    result := sl.IndexOfObject(ci); 
end;


procedure TClipQueue.GetQueueItems(items : TStrings);
begin
    items.Clear;
    items.AddStrings( sl );
end;


function TClipQueue.GetClipItem(index: cardinal): TClipItem;
    var ci : TClipItem;
begin
    ci := TClipItem(sl.Objects[index]);
    result := ci;
end;




//
// Move items in the list
//
procedure TClipQueue.MoveToStart(index : cardinal);
begin
    sl.Move(index, 0);
end;

procedure TClipQueue.Move(oldIndex : cardinal; newIndex : cardinal);
begin
    sl.Move(oldIndex, newindex);
end;



//
// Remove items from queue
// NOTE: Resources must be released here
//
procedure TClipQueue.DeleteItem(index : cardinal);
var ci : TClipItem;
begin

    // We've got to clean up the clone icon to avoid a memory leak
    ci := TClipITem(sl.Objects[index]);

    // Add the text to the removed items
    // ClipItem also has global memory to free
    RemovedQueue.InsertAtStart(self.GetItemText(index), ci);
    TempClipQueue.InsertAtStart(ci);
    {
    if (ci <> nil) then begin
        UnitMisc.AppendLog('^Deleting Icon^');
        MyDestroyIcon(ci.CData.GetHICON);
    end;
    MyFree(ci);
    }

    // this MUST be the ONLY place an item is deleted
    sl.Delete(index);
end;


procedure TClipQueue.ClearQueue;
var i : longint;
begin
    // see note in DeleteQueueItem for rules of removing items
    // from queue
    // I know better, this must go in reverse order
    for i := (sl.count - 1) downto 0 do begin
        self.DeleteItem(i);
    end;

end;













//------------------------------
// ClipData / ClipItem
//------------------------------

{ TClipData }

constructor TClipData.Create(s: string; h: HICON);
begin
    self.s := s;
    self.h := h;
end;

destructor TClipData.Destroy;
var s : string;
begin
    s := 'du what?';
  inherited;
end;

function TClipData.GetCreationDate: TDateTime;
begin
    result := self.timestamp;
end;

function TClipData.GetHICON: HICON;
begin
    if (self.h = 0) then begin
        result := UnitClipQueue.ClipDataDefaultIcon;
    end else begin
        result := self.h;
    end;
end;

function TClipData.GetHICONAbsolute: HICON;
begin
    result := self.h;
end;

function TClipData.GetString: string;
begin
    result := self.s;
end;

procedure TClipData.SetHICON(h: HICON);
begin
    //if h <> 0 then DestroyIcon(h);
    
    self.h := h;
end;

procedure TClipData.SetString(s: string);
begin
    self.s := s;
end;

{ TClipItem }

constructor TClipItem.Create;
begin
    self.CData := TClipData.Create('', 0);
    self.CHandle := 0;
    self.CFormat := 0;
end;

destructor TClipItem.Destroy;
begin
    self.CleanupMemory;
    MyFree(CData);
    inherited Destroy;
end;

procedure TClipItem.CleanupMemory;
begin
    if (self.CHandle <> 0) then begin
        Windows.GlobalFree(self.CHandle);
        self.CHandle := 0;
    end;
end;




function TClipItem.GetAsText: string;
begin
    result := CData.GetString;
end;

//
// Return the format of clipitem saved
// 0 = ERROR
// NOTES: GetAtText will return empty unless the clipboard has CF_TEXT or
// CF_HDROP
//
function TClipItem.GetClipboardItem( hi: HICON; OverrideFormat : word = 0; SizeLimit : cardinal = $FFFF): cardinal;
var files : string;
    CVolatileHandle : THandle;
    IsText : boolean;
    HasText : boolean;
    Procedure ChooseAFormat(backup : boolean = false);
    var i : integer;
    begin

        if (not Windows.OpenClipboard(Application.Handle)) then begin
            UnitMisc.AppendLog('<ClipItem - can''t open clipboard> ', true);
            EXIT;
        end;


        // This gets the first and
        // most descriptive clipboard format the clipboard contains
        self.CFormat := Windows.EnumClipboardFormats(0);
        if backup then begin
            UnitMisc.AppendLog('format=' + Self.GetFormatName(false));
            self.CFormat := Windows.EnumClipboardFormats(self.CFormat);
        end;
        Windows.CloseClipboard;


        // OVERIDE format
        // CF_BITMAP kills over in DupHandle, CD_DIB does not
        // A wave clip may show up as a speaker icon if CF_WAVE is not
        // before CF_DIB

        HasText := Clipboard.HasFormat(CF_TEXT);
        if clipboard.HasFormat(CF_WAVE) then begin
            CFormat := CF_WAVE
        {end else if clipboard.HasFormat(CF_BITMAP) then begin
            CFormat := CF_BITMAP;}
        end else if clipboard.HasFormat(CF_DIB) then begin
            CFormat := CF_DIB
        end else if clipboard.HasFormat(CF_HDROP) then begin
            CFormat := CF_HDROP;
        end;
        { else if clipboard.HasFormat(UnitMisc.GetCF_RICHTEXT) then begin
            CFormat := UnitMisc.GetCF_RICHTEXT;
            IsText := true;
        end else if clipboard.HasFormat(UnitMisc.getCF_HTML) then begin
            CFormat := UnitMisc.GetCF_HTML;
            IsText := true;
        end else if clipboard.HasFormat(CF_UNICODETEXT) then begin
            CFormat := CF_UNICODETEXT;
            IsText := true;
        end; }

        for i := 0 to FrmConfig.FormatCount - 1 do begin
            case FrmConfig.GetFormat(i) of
            FO_UNICODE :
                begin
                    if clipboard.HasFormat(CF_UNICODETEXT) then begin
                        CFormat := CF_UNICODETEXT;
                        IsText := true;
                        BREAK;
                    end;
                end;
            FO_RICHTEXT:
                begin
                    if clipboard.HasFormat(UnitMisc.GetCF_RICHTEXT) then begin
                        CFormat := UnitMisc.GetCF_RICHTEXT;
                        IsText := true;
                        BREAK;
                    end
                end;
            FO_HTML:
                begin
                    if clipboard.HasFormat(UnitMisc.getCF_HTML) then begin
                        CFormat := UnitMisc.GetCF_HTML;
                        IsText := true;
                        BREAK;
                    end
                end;
            end;
        end;

        if (CFormat = CF_OEMTEXT) or
            ((self.CFormat = 0) and HasText) then begin
            CFormat := CF_TEXT;
            IsText := true;
        end;
        UnitMisc.AppendLog('<ClipItem Overided Format = ' + self.GetFormatName(false) );
    end;
    procedure LastDitchAttempt;
    begin
        if HasText then begin
            CData.SetString(Clipboard.AsText);
        end;
    end;

var Estimation : cardinal;
begin
    UnitMisc.AppendLog('<ClipItem> ', true);
    Windows.SetLastError(ERROR_SUCCESS);     // workaround for some OS's
    CData.SetHICON(HI);  // incase retreival fails, we don't want to leak memory


    // find the current format
    // must be non-zero
    self.CleanupMemory;
    CFormat := 0;
    CHandle := 0;
    result := 0;
    IsText := false;
    try
        if OverrideFormat = 0 then begin
            ChooseAFormat;
        end else begin
            CFormat := OverrideFormat;
        end;

    except
        UnitMisc.AppendLog('<ClipItem - Find Format exception' + SysErrorMessage(GetLastError) );
        Windows.CloseClipboard;
        result := 0;
        CFormat := 0;
        CHandle := 0;
        EXIT;
    end;

    if (CFormat  = 0) then begin
        EXIT;
    end;


    //
    // Dupe the handle, exit on error
    //
    try

        //
        // Win9X will puke if you try to dup the handle and the clipboard is
        // not open. The VolatileHandle name reminds me that it is not to be
        // touched anywhere but here.
        //
        self.CData.size := SizeLimit;
        if not Windows.OpenClipboard(Application.Handle) then begin
            UnitMisc.AppendLog('<ClipItem - can''t open clipboard2 > ', true);
            LastDitchAttempt;
            EXIT;
        end;



        CVolatileHandle := Windows.GetClipboardData(CFormat);
        if CVolatileHandle = 0 then begin
            Windows.CloseClipboard;
            UnitMisc.AppendLog('<ClipItem - can''t get handle > using backup format', true);
            ChooseAFormat(true);
            if not Windows.OpenClipboard(Application.Handle) then begin
                UnitMisc.AppendLog('<ClipItem - can''t open clipboard3 > ', true);
                EXIT;
            end;


            CVolatileHandle := Windows.GetClipboardData(CFormat)
        end;
        if (CVolatileHandle = 0) then begin
            UnitMisc.AppendLog('<ClipItem - can''t get handle > ', true);

            If HasText then begin
                CFormat := CF_TEXT;
                CVolatileHandle := Windows.GetClipboardData(CFormat);
                if (CVolatileHandle = 0) then begin
                    UnitMisc.AppendLog('<ClipItem - can''t get handle > ', true);
                    Windows.CloseClipboard;
                    EXIT;
                end;
                UnitMisc.AppendLog('<Clip Item - getting plain text>');
            end;
        end;
        CHandle := UnitMisc.DupHandle(CVolatileHandle, self.CData.size, (SizeLimit <> $FFFF));
        if (CHandle = 0) then begin
            UnitMisc.AppendLog('<ClipItem - can''t dup handle> ');
            Windows.CloseClipboard;
            LastDitchAttempt;
            EXIT;
        end;
        Windows.CloseClipboard;

        // save the extra data
        // Icon handle, Text version of item

        CData.SetHICON(HI);

        if (IsText or HasText) then begin
            CData.SetString(Clipboard.AsText);

            // vba workaround - garbage in the clip
            if (CFormat = CF_UNICODETEXT) and (cdata.size = 2048) then begin
                Estimation := ((length(cdata.s)+1)*2);
                if Estimation <> Cdata.size  then begin
                    cdata.size := Estimation;
                end;

            end;

        end else if (CFormat = CF_HDROP) then begin;
            files := self.GetFilenamesAsText(CHandle) ;
            UnitMisc.AppendLog('filenames: ' + files);
            CData.SetString(files);
        end else begin
            CData.SetString('');
        end;



    except
        on E: Exception do begin
            UnitMisc.AppendLog('<clipItem Dup Exception - ' + E.Message + ' ', true);
            Windows.CloseClipboard;
            CFormat := 0;
            CHandle := 0;
            result := 0;
            EXIT;
        end;
    end;

    {Future user - when the clip was created}
    cdata.timestamp := now;
    UnitMisc.AppendLog('<ClipItem ClipboardSave success!> size=' + IntToSTr(CData.size) );
    result := CFormat;
end;


// return all filenames or return empty string
function TClipItem.GetFilenamesAsText(h : THandle): string;
var i, j : longint;
    s : string;
begin
    Windows.SetLastError(ERROR_SUCCESS);
    result := '';

    if (CFormat = CF_HDROP) then begin
        UnitMisc.AppendLog('    Detecting Filenames...');

        try
            //
            // On Win9x, the handle must be on the clipboard and the clipboard
            // must be open; otherwise, bad bad stuff happens
            //
            UnitMisc.AppendLog('GetFilenamesAsText');
            s := stringofchar(#0, Windows.MAX_PATH+1);
            j := ShellApi.DragQueryFile(h, Cardinal(-1){//$FFFFFFFF}, nil, 0);
            UnitMisc.AppendLog('FileCount=' + IntToSTr(j));
            result := '';
            for i := 0 to (j - 1) do begin
                s := stringofchar(#0, Windows.MAX_PATH+1);
                ShellApi.DragQueryFile(h, i, pchar(s), Windows.MAX_PATH);
                UnitMisc.AppendLog(Trim(String(s)));
                if i = (j - 1) then begin
                    result := result + Trim(PChar(s));
                end else begin
                    result := result + Trim(PChar(s)) + #13 + #10;
                end;
            end;

        except
            on e : exception do begin
                UnitMisc.AppendLog('GetFilenamesAsText' + #13#10 +
                    e.Message + ' ', true
                );
            end;
        end;
    end;
end;

function TClipItem.GetFormat: cardinal;
begin
    result := self.CFormat;
end;

function TClipItem.GetHandle: THandle;
begin
    result := self.CHandle;
end;

procedure TClipItem.OverrideTextVersionOfItem(s : string);
begin
    if self.CFormat = 0 then self.CFormat := CF_TEXT;
    self.CData.SetString(s);
end;

{
procedure TClipItem.PlaceOnClipboard;
begin
    // moved all pasting logic to the appropreite location, the Paste Object
    if (CHandle <> 0) then begin
        Paste.SetClipboardOnlyOnce;
        Paste.SendText('', self);
    end else begin
        UnitMisc.AppendLog('ClipItem - PlaceOnClipboard failed ' + SysErrorMessage(GetLastError));
    end;
end;
}
function TClipItem.GetFormatName(AccessHandle : boolean = true) : string;
var name : array[0 .. 80] of char;
    format : cardinal;
    p : ^tagBITMAPINFO;
begin
    Windows.SetLastError(ERROR_SUCCESS);

    format := self.CFormat;

    case (format) of
        CF_DIB          : begin
            if (AccessHandle) then begin
                p := Windows.GlobalLock(CHandle);
                if (p <> nil) then begin
                    result := 'Picture (DIB) ' + IntToStr(p^.bmiHeader.biWidth) + 'x' + IntToStr(p^.bmiHeader.biHeight);
                    Windows.GlobalUnlock(self.Chandle);
                end else begin
                    UnitMisc.AppendLog('GetFormatName: ' + SysErrorMessage(GetLastError));
                end;
            end else begin
                result := 'Picture (DIB)';
            end;
        end;
        CF_WAVE         : result := 'Wave Audio';
        CF_BITMAP       : result := 'Picture (Bitmap)';
        CF_HDROP        : result := 'File(s)';

        CF_DIF          : result := 'CF_DIF';
        CF_TEXT         : result := 'Plain text';
        CF_SYLK         : result := 'CF_SYLK';
        CF_TIFF         : result := 'CF_TIFF';
        CF_RIFF         : result := 'CF_RIFF';
        CF_LOCALE       : result := 'CF_LOCALE';
        CF_OEMTEXT      : result := 'CF_OEMTEXT';
        CF_PALETTE      : result := 'CF_PALETTE';
        CF_PENDATA      : result := 'CF_PENDATA';
        CF_UNICODETEXT  : result := 'Unicode';
        CF_ENHMETAFILE  : result := 'CF_ENHMETAFILE';
        CF_METAFILEPICT : result := 'CF_METAFILEPICT';
        else begin
            if format = UnitMisc.GetCF_HTML then begin
                result := 'HTML';
            end else if format = UnitMisc.GetCF_RICHTEXT then begin
                result := 'RichText';
            end else begin
                UnitMisc.AppendLog('Unknown format...');
                if Windows.GetClipboardFormatName(format,  @name,  sizeof(name)  ) <> 0 then begin
                    Result := string(name);
                    if (result  = '') then begin
                        result := 'Unknown';
                    end;
                end else begin
                    UnitMisc.AppendLog('Format Name failed', true);
                    result := 'Error: Unknown';
                end;
            end;
        end;
   end;

end;

procedure TClipItem.GetDIB( pic : TPicture);
    procedure DibToBitmap(hDIB:  THandle; BM : TBitmap);
    var
        bmfh :  TBitmapFileHeader;
        bi :  PBitmapInfo;
        ms :  TMemoryStream;
        ColorsUsed : integer;

    begin
        bi := PBitmapInfo( Windows.GlobalLock(hDIB) );
        if bi = nil then begin
            UnitMisc.AppendLog('DibToBitmap - GlobalLock', true);
            EXIT;
        end;
        try
            // Caculate the number of colors used (power of 2)
            ColorsUsed := bi.bmiHeader.biClrUsed;
            if (ColorsUsed = 0) and (bi.bmiHeader.biBitCount <= 8) then
                ColorsUsed := 1 shl bi.bmiHeader.biBitCount;

            bmfh.bfType := $4D42;  // 'BM'
            bmfh.bfReserved1 := 0;
            bmfh.bfReserved2 := 0;

            // point to location of actual data
            // header is variable because of the RGBQuads
            bmfh.bfOffBits := SizeOf(TBitmapFileHeader)       +
                       SizeOf(TBitmapInfoHeader)       +
                       ColorsUsed * SizeOf(TRGBQuad);
            // from msdn: DIBs and Their Use (GDI Technical Articles)
            //  "The crazy roundoffs and shifts account for the
            //  bitmap being DWORD-aligned at the end of every scanline.""
            // Hooray for microkernel garbage!  
            if bi.bmiHeader.biSizeImage = 0 then begin
                with bi.bmiHeader do begin
                    biSizeImage :=
                    ((((biWidth * biBitCount) + 31) and
                        (not 31)) shr 3) * biHeight;
                end;
            end;
            bmfh.bfSize := bmfh.bfOffBits + bi.bmiHeader.biSizeImage;

            // Create a fake bitmap file
            // and load it into a TBitmap
            ms := TMemoryStream.Create;
            try
                ms.Write(bmfh, SizeOf(TBitmapFileHeader));
                ms.Write(bi^, bmfh.bfSize - SizeOf(TBitmapFileHeader));
                ms.Position := 0;

                BM.LoadFromStream(ms)
            finally
                ms.Free
            end;
        finally
            Windows.GlobalUnlock(hDIB);
        end;
    end;

begin
    Windows.SetLastError(ERROR_SUCCESS);

    if (self.CFormat = CF_DIB) then begin
        DibToBitmap(self.CHandle, pic.Bitmap);
    end;
end;

procedure TClipItem.GetRichText(var s: string);
var p : PChar;
    ss : TStringStream;
begin
    p := Windows.GlobalLock(self.CHandle);
    if (p <> nil) then begin
        ss := TStringStream.Create('');
        ss.Write(p^, Windows.GlobalSize(self.CHandle));
        s := ss.DataString;

        Windows.GlobalUnlock(self.Chandle);

        MyFree(ss);
    end;
end;


//
// Static Objects
//

function TClipItem.GetDataSize: cardinal;
begin
    result := self.CData.size;
end;

function TClipItem.GetFilename(path: string; index: integer): string;
begin
    result := path + IntToHex(index, 8) + FILE_EXT_BINARY;
end;


procedure TClipItem.SaveToFile(path : string; index : integer);
var f : file;
    tf : textfile;
    p : Pointer;
    w : word;
    clpname, s : string;
begin
    if self.CHandle = 0 then EXIT;
    if self.CData.size = 0 then EXIT;

    clpname := self.GetFilename(path, index);
    assign(f, clpname);
    rewrite(f,1);

    if (self.CFormat = UnitMisc.GetCF_RICHTEXT) then begin
        w := CF_FILE_RICHTEXT;
    end else if (self.CFormat = UnitMisc.GetCF_HTML ) then begin
        w := CF_FILE_HTML;
    end else begin
        w := self.CFormat;
    end;
    blockwrite(f, w, sizeof(self.CFormat));
    blockwrite(f, self.CData.size, sizeof(self.CData.size));

    p := GlobalLock(self.CHandle);
    blockwrite(f, p^, self.cdata.size);
    GlobalUnlock(self.CHandle);

    closefile(f);

    // simple text version
    s := self.GetAsText;
    if s = '' then begin
        EXIT;
    end;

    clpname := self.GetFilename(path, index);
    clpname := leftstr(clpname, length(clpname) - 4) + FILE_EXT;
    assign(tf, clpname);
    rewrite(tf);

    write(tf,s);
    closefile(tf);
end;

procedure TClipItem.SetFromWideChar(wc: TWideChar);
var p : pointer;
begin
    self.CFormat := CF_UNICODETEXT;


    self.CHandle := GlobalAlloc(GMEM_MOVEABLE, wc.size);
    p := GlobalLock(self.CHandle);
    if (p = nil) then begin
        self.CHandle := 0;
        self.CData.size := 0;
        UnitMisc.AppendLog('TClipItem.LoadFromFile: error couldn''t lock handle!');
        EXIT;
    end;

    try
        wc.GetMemoryStream.Read(P^, wc.size);
        //move(wc.memory^, p^, wc.size);
        GlobalUnlock(self.CHandle);
    except
        on e: exception do begin
            self.CHandle := 0;
            self.CData.size := 0;
            UnitMisc.AppendLog('TClipItem.LoadFromFile: read2 error' + SysErrorMessage(GetLastError));
        end;
    end;
end;

procedure TClipItem.LoadFromFile(path : string; index : integer);
var f : file;
    p : pointer;
    clpname, s : string;
    i : integer;
label EXIT_CODE;
begin
    clpname := self.GetFilename(path, index);
    if not (FileExists(clpname)) then begin
        EXIT;
    end;

    assignfile(f, clpname);
    reset(f,1);
    if filesize(f) = 0 then begin
        goto EXIT_CODE;
    end;
    try
        blockread(f, self.CFormat, sizeof(self.CFormat));
        if (self.CFormat =  CF_FILE_RICHTEXT) then begin
            self.CFormat := UnitMisc.GetCF_RICHTEXT;
        end else if (self.CFormat = CF_FILE_HTML) then begin
            self.CFormat := UnitMisc.GetCF_HTML;
        end;
        blockread(f, self.CData.size, sizeof(self.CData.size));
    except
        on e: exception do begin
            self.CHandle := 0;
            self.CData.size := 0;
            UnitMisc.AppendLog('TClipItem.LoadFromFile: read1 error' + SysErrorMessage(GetLastError) + e.Message );
            closefile(f);
            exit;
        end;
    end;

    self.CHandle := GlobalAlloc(GMEM_MOVEABLE,self.CData.size);
    p := GlobalLock(self.CHandle);
    if (p = nil) then begin
        self.CHandle := 0;
        self.CData.size := 0;
        UnitMisc.AppendLog('TClipItem.LoadFromFile: error couldn''t lock handle!');
        goto EXIT_CODE;
    end;

    try
        blockread(F, p^, self.CData.size);
        GlobalUnlock(self.CHandle);
    except
        on e: exception do begin
            self.CHandle := 0;
            self.CData.size := 0;
            UnitMisc.AppendLog('TClipItem.LoadFromFile: read2 error' + SysErrorMessage(GetLastError));
        end;
    end;
    closefile(f);


    clpname := self.GetFilename(path, index);
    clpname := leftstr(clpname, length(clpname) - 4) + FILE_EXT;
    if not fileexists(clpname) then EXIT;
    assign(f, clpname);
    reset(f, 1);
    i := filesize(f);
    if i <> 0 then begin
        try
            setlength(s, i);
            blockread(f, s[1], i);
            self.CData.SetString( s );
        except
        end;
    end;

EXIT_CODE:
    closefile(f);
end;

    {
    function BitmapToIcon(Bitmap: TBitmap): HICON;
    var x, y : integer;
        info : TIconInfo;
        image, mask : TBitmap;
        i, j : Integer;
        TransparentColor : TColor;
    begin
        // this entire routine doesn't work - it erases black instead
        // of preserving tranparency

        X := 32;
        Y := 32;

        image:= TBitmap.Create;
        image.Width:= X;
        image.Height:= Y;

        image.Canvas.StretchDraw(Rect(0, 0, X, Y), Bitmap);
        image.TransparentColor:= Bitmap.TransparentColor;

        TransparentColor:= image.TransparentColor and $FFFFFF;

        mask:= TBitmap.Create;
        //mask.Assign(image);
        mask.Monochrome := true;
        mask.Width := image.Width;
        mask.Height := image.Height;

        for i := 0 to (Y - 1) do begin
            for j:= 0 to (X - 1) do begin
                if (image.Canvas.Pixels[i, j] = TransparentColor) then begin
                    mask.Canvas.Pixels[i, j]:= clWhite;
                end else begin
                    mask.Canvas.Pixels[i, j]:= clBlack;
                end;
                //image.Canvas.Pixels[i,j] := image.Canvas.Pixels[i,j];
            end;
        end;

        info.fIcon := True;
        info.hbmMask := mask.MaskHandle;
        info.hbmColor := image.Handle;

        Result := CreateIconIndirect(info);

        mask.Free;
        image.Free;
    end;
    }


function TClipItem.GetIconFilename(path: string; index: integer; sufix : string = ''): string;
begin
    result := path + IntToHex(index, 8) + sufix + '.bmp';
end;

procedure TClipItem.LoadIconFromFile(path : string; index : integer; sufix : string = '');
var icn : TIcon;
    bit1, bit2 : TBitmap;
    info : _ICONINFO;
    iconName : string;
    h : HICON;
begin
    // legacy support
    // load the ICO version only if the BMP
    // version does not exist

    iconName := self.GetIconFilename(path, index, sufix);

    if (FileExists( stringreplace(iconName,'.bmp','-m.bmp',[rfignorecase])))
        and (FileExists( stringreplace(iconName,'.bmp','-c.bmp',[rfignorecase]))) then begin
    {if (FileExists(IconName)) then begin}
        // preserve transparency
        try
            bit1 := TBitmap.Create;
            bit1.LoadFromFile(stringreplace(iconName,'.bmp','-m.bmp',[rfignorecase]));
            bit2 := TBitmap.Create;
            bit2.LoadFromFile(stringreplace(iconName,'.bmp','-c.bmp',[rfignorecase]));

            info.fIcon := true;
            info.xHotspot := 0;
            info.yHotspot := 0;
            info.hbmMask := bit1.Handle;
            info.hbmColor := bit2.Handle;

            h := Windows.CreateIconIndirect(info);
            self.CData.SetHICON( h );
            MyFree(bit1);
            MyFree(bit2);
        except
        end;
        {
        bit := TBitmap.Create;
        bit.LoadFromFile(IconName);
        self.CData.SetHICON( BitmapToIcon(bit) );
        MyFree(bit);
        }
    end else begin
        iconName := path + IntToStr(index) + sufix + '.ico';

        if (FileExists(iconName)) then begin
            {assignfile(f, iconname);
            Reset(f);
            if (FileSize(f) <> 0) then begin
                CloseFile(f);}
            try
                icn := TIcon.Create;
                icn.LoadFromFile(iconname);

                self.CData.SetHICON(icn.Handle);
            except
                // ignore empty / null icon files
            end;
                //icn.free; {Can't free this icon - it will destroy the HICON}
            //end;
        end;
    end;
end;

procedure TClipItem.SaveIconToFile(path : string; index : integer; sufix : string = '');
var bit1, bit2 : TBitmap;
    info : _ICONINFO;
    iconName : string;
begin
    iconName := Self.GeticonFilename(path, index, sufix);
    if (self.CData.GetHICONAbsolute <> 0) then begin
        bit1 := TBitmap.Create;
        bit2 := TBitmap.Create;

        Windows.GetIconInfo(self.CData.GetHICONAbsolute, info);
        bit1.Handle := info.hbmMask;
        bit1.SaveToFile( stringreplace(iconName,'.bmp','-m.bmp',[rfIgnoreCase]));

        bit2.Handle := info.hbmColor;
        bit2.SaveToFile( stringreplace(iconName,'.bmp','-c.bmp',[rfIgnoreCase]));
        MyFree(bit1);
        MyFree(bit2);

    end else begin
        //
        // Without this, null icons will use old cached icon from other clips
        //
        DeleteFile(stringreplace(iconName,'.bmp','-m.bmp',[rfIgnoreCase]));
        DeleteFile(stringreplace(iconName,'.bmp','-c.bmp',[rfIgnoreCase]));
    end;
    {
    //
    // This works, but I can't find a way to load it correctly.
    // Saving mask and image separately works for me
    //
    bit := TBitmap.Create;
    bit.width := 32;
    bit.height := 32;
    // 24 bit kills transparency
    bit.PixelFormat := pf8bit;

    DrawIcon(bit.Canvas.Handle,0,0,self.CData.GetHICON);

    bit.SaveToFile(iconName);
    MyFree(bit);
    }
end;

function TClipItem.HasText: boolean;
begin
    result := (self.GetAsText <> '') and (self.CFormat <> CF_HDROP);
end;



{ TPagedStringQueue }

//--------------------
// Public Interface
//--------------------
constructor TPagedStringQueue.Create(filename: string; folder : string = '');
var tf : textfile;
    s, itemText : string;
    linecount, i,j, itemcount : integer;
begin
    self.ci := TClipItem.Create;
    iconcachecrc := THashedStringList.create;

    self.base := UnitMisc.GetAppPath; 
    if folder = '' then begin
        self.cache :=  self.base + 'removedcache\';
    end else begin
        self.cache := IncludeTrailingPathDelimiter( self.base + folder );
    end;
    self.filename := self.base + filename;
    SysUtils.ForceDirectories(cache);
    self.iconcache := self.cache + 'iconcache\';
    SysUtils.ForceDirectories(self.iconcache);

    //
    // legacy stuff - move the old file into the new cache
    //
    if (FileExists(self.filename)) then begin
        AssignFile(tf, filename);
        Reset(tf, filename);

        try
            itemcount := 0;
            while not eof(tf) do begin
                try
                    Readln(tf, s);
                    itemText := '';
                    lineCount := StrToInt(s);

                    for i := 0 to lineCount - 1 do begin
                        Readln(tf, s);
                        if (itemText = '') then begin
                            itemText := s;
                        end else begin
                            itemText := itemText + #13#10 + s;
                        end;
                    end;
                    self.qSize := itemCount + 1;
                    self.SaveItem(itemText, itemcount);

                    inc(itemcount);
                except
                    on E: Exception do begin
                        ShowMessage('The "Load items from last session" file is corrupted - ' + filename + #13#10#13#10 +
                                    'Error Message: ' + E.Message);
                        break;
                    end;
                end;
            end;
        finally
            CloseFile(tf);
        end;

        DeleteFile(self.filename);
    end;


    //
    // convert the old format into the new format
    //
    i := 0;
    while
        FileExists(self.GetFilenameAbsoluteOld(i))
    do begin
        if fileexists(self.GetFilenameAbsolute(i)) then begin
            deletefile(self.GetFilenameAbsolute(i));
        end;
        renamefile(
            self.GetFilenameAbsoluteOld(i),
            self.GetFilenameAbsolute(i)
        );
        inc(i);
    end;



    // find the queue size by counting contiguous files
    // get the real "zero" item index
    // count the contiguous files, starting at qStart for the qcount

    self.qSize := 0;
    i := 0;
    self.qSize := $ffffffff;
    while
        FileExists(self.GetFilenameAbsolute(i))
    do begin
        inc(i);
    end;
    self.qSize := i;


    if (FileExists(self.cache + PAGED_STATEFILE)) then begin
        assignfile(tf, self.cache + PAGED_STATEFILE);
        reset(tf);
        read(tf, self.qStart);
        close(tf);
    end else begin
        self.qStart := 0;
    end;

    self.qCount := 0;
    if self.qStart > qSize then begin
        // corrupt cache, what to do?
        self.qStart := 0;
    end else begin
        for j := qstart to qsize - 1 do begin
            if (self.IsEmptyItemAbsolute(j)) then EXIT;
            inc(self.qCount);
        end;
        for j := 0 to qstart-1 do begin
            if (self.IsEmptyItemAbsolute(j)) then EXIT;
            inc(self.qCount);
        end;
    end;

end;

destructor TPagedStringQueue.Destroy;
begin
    inherited;
end;

procedure TPagedStringQueue.InsertAtStart(s: string; ci : TClipItem = nil);
begin
    // ensure the queue has 1 free spot
    while (self.qCount >= self.qSize) and (self.qCount <> 0) do begin
        self.RemoveOldestItem;
    end;

    // save the item at the end of the queue
    self.SaveItem(s, self.qCount, ci);
    inc(self.qCount);
    self.SaveQueueState;

    {
    while (self.qCount >= self.qSize) and (self.qCount <> 0) do begin
        self.RemoveOldestItem;
    end;
    }
end;

function TPagedStringQueue.GetQueueCount: cardinal;
begin
    result := self.qCount;
end;

procedure TPagedStringQueue.ClearQueue;
var fn : string;
    tf : textfile;
    i : Integer;
begin
    // no more dirty tricks
    // - empty every placeholder items
    // - ensure no file exists after the last item
    for i := 0 to (self.qsize - 1) do begin
        fn := self.GetFilenameAbsolute(i);
        Assign(tf, fn);
        Rewrite(tf);
        Close(tf);
    end;

    fn := self.GetFilenameAbsolute(self.qsize);
    if (FileExists(fn)) then begin
        DeleteFile(fn);
    end;

    self.qStart := 0;
    self.qCount := 0;
    self.SaveQueueState;
end;

procedure TPagedStringQueue.SaveQueueState;
var tf : textfile;
begin
    assignfile(tf, cache + PAGED_STATEFILE);
    rewrite(tf);
    write(tf, self.qStart);
    close(tf);
end;


procedure TPagedStringQueue.SetQueueSize(size: cardinal);
var fn : string;
    i : integer;
    tf : textfile;
begin
    if self.qSize = size then EXIT;


    // Make sure all files exist within the circular queue
    // (when the queue size is enlarged)
    // Make sure new items are blank

    for i := 0 to (size - 1) do begin
        fn := self.GetFilenameAbsolute(i);
        if not FileExists(fn) or
            (cardinal(i) >= self.qSize) then begin
            Assign(tf, fn);
            Rewrite(tf);
            Close(tf);
        end;
    end;


    // To Shrink the paged queue.....
    // Remove oldest items
    // Re-index items from 0-(qSize - 1)  so that virtual indexes
    // match the actuall indexes.
    // Break the contigous numbers so that only 0-(size-1) exists
    // contigously

    while (self.qCount > size) do begin
        self.RemoveOldestItem;
    end;

    // investigate - do we ever need to re-index the queue?
    //self.ReIndexQueue;

    fn := self.GetFilenameAbsolute(size);
    if (FileExists(fn)) then begin
        DeleteFile(fn);
    end;

    self.qSize := size;
    self.SaveQueueState;

end;

function TPagedStringQueue.GetItemText(index: cardinal): string;
begin
    index := IndexTranslate(index);
    result := self.GetItemAbsolute(index);
end;
function TPagedStringQueue.GetItemClip(Index: cardinal): TClipItem;
var fs : TFileStream;
    c : cardinal;
    ti : integer;
    s : string;
    count : integer;
begin
    ti := IndexTranslate(index);
    result := self.GetItemClipAbsolute(ti);
    if (result <> nil) then begin
        //
        // this causes icons with the same CRC32 to use
        // the re-read the same icon file - exploiting the OSs file system cache
        //
        fs := TFileStream.Create(self.iconcache + 'iconindex.dat', fmCreate or fmOpenRead);
        fs.Seek(ti, soFromBeginning);
        count := fs.read(c,sizeof(c));

        // duplicate icon, or new icon found
        if count <> 0 then begin
            s := iconcachecrc[c];
            if (s <> '') then begin
                ti := StrToInt(s);
            end else begin
                iconcachecrc[c] := IntToStr(ti);
            end;
        end;
        result.LoadIconFromFile(self.iconcache,  ti);
        myfree(fs);
    end;
end;

function TPagedStringQueue.GetItemDate(index: cardinal): TDateTime;
var f : string;
begin
    index := IndexTranslate(index);

    f := self.GetFilenameAbsolute(index);
    result := FileDateToDateTime(FileAge(f));
end;




//-----------------------
// Private Implementation
//-----------------------

procedure TPagedStringQueue.SaveItem(s : string; index: cardinal; ci : TClipItem = nil);
var b : TBitmap;
    ms : TMemoryStream;
    fs : TFileStream;
    crc : TCRC32;
    c : cardinal;
    info : _ICONINFO;
    buff : array of word;
    fn : string;
    i : integer;
begin
    if self.qSize = 0 then EXIT;
    if (ci <> nil) then begin
        ci.SaveToFile(self.cache, self.IndexTranslate(index) );
        if (ci.CData.GetHICONAbsolute <> 0) then begin
            ci.SaveIconToFile(self.iconcache, self.IndexTranslate(index) );

            b := TBitmap.Create;
            ms := TMemoryStream.create;
            crc := TCRC32.Create;
            // convert icon to buffer
            Windows.GetIconInfo(ci.CData.GetHICONAbsolute, info);
            b.Handle := info.hbmColor;
            b.SaveToStream(ms);
            SetLength(buff, ms.size div 2);
            ms.Seek(0, soFromBeginning);
            ms.Read(buff[0], ms.size);

            // buffer to CRC to index in file
            for I := low(buff) to high(buff) do begin
                c := crc.Update(buff[i]);
            end;

            fn := self.iconcache + 'iconindex.dat';
            fs := TFileStream.Create(fn, fmCreate or fmOpenWrite);
            fs.Seek(self.IndexTranslate(index),soFromBeginning);
            fs.Write(c, sizeof(c));

            myfree(fs);
            myfree(crc);
            myfree(ms);
            myfree(b);
        end;
    end;
end;


function TPagedStringQueue.IsEmptyItem(index: cardinal): boolean;
begin
    result := self.IsEmptyItemAbsolute(IndexTranslate(index));
end;
function TPagedStringQueue.IsEmptyItemAbsolute(
  absoluteIndex: cardinal): boolean;
var f : file;
    fn, newfn : string;
    fs : cardinal;
begin
    result := true;
    newfn := self.GetFilenameAbsolute(absoluteIndex);
    fn := self.GetFilenameAbsoluteOld(absoluteIndex);
    if fileexists(newfn) then begin
        fn := newfn;
    end;

    Assign(f, fn);
    Reset(f, 1);
    fs := 0;
    try
        fs := FileSize(f);
    finally
        result := (fs = 0);
        Close(f);
    end;
end;

function TPagedStringQueue.IsEmptyItemClipAbsolute(
  absoluteIndex: cardinal): boolean;
var f : file;
    fn : string;
    fs : cardinal;
begin
    fn := self.GetFilenameClipAbsolute(absoluteIndex);
    Assign(f, fn);
    Reset(f, 1);
    try
        fs := FileSize(f);
        result := (fs = 0);
    finally
        Close(f);
    end;
end;




function TPagedStringQueue.IndexTranslate(index: cardinal): cardinal;
begin
    // since this is a circular queue, item "0" can actually
    // start anywhere from 0 to qSize - 1

    result := (self.qStart + index) mod self.qSize;
end;


procedure TPagedStringQueue.RemoveOldestItem;
var fn : string;
    tf : textfile;
begin
    // "delete" the first/oldest item in the list
    // empty files must exist to keep a contiguous run of files

    if (self.qSize = 0) then begin
        fn := self.GetFilenameAbsolute(0);
        if (FileExists(fn)) then begin
            Assign(tf, fn);
            Rewrite(tf);
            Close(tf);
        end;
        self.qCount := 0;
    end else begin
        self.SaveItem('', 0);
        self.qStart := (self.qStart + 1) mod qSize;
        dec(self.qCount);
    end;
end;

{procedure TPagedStringQueue.ReIndexQueue;
var s1, s2 : string;
    i, k : integer;
const BAK_EXT = '.bak';
begin
    if self.qStart = 0 then EXIT;

    // move virtual items 0-n to absolute indexes 0-n
    // replace the extension of all items so there are no name clashed
    // when re-ordering the items

    for i := 0 to self.qSize-1 do begin
        s1 := self.GetFilenameAbsolute(i);
        RenameFile(s1, s1 + BAK_EXT);



        s1 := self.GetFilenameClipAbsolute(i);
        RenameFile(s1, s1 + BAK_EXT);

        s1 := self.GetFilenameClipAbsolute(i);
        s1 := leftstr(s1,length(s1)-4) + FILE_EXT;
        RenameFile(s1, s1 + BAK_EXT);


        s1 := self.GetFilenameIcon1Absolute(i);
        RenameFile(s1, s1 + BAK_EXT);

        s1 := self.GetFilenameIcon2Absolute(i);
        RenameFile(s1, s1 + BAK_EXT);
    end;


    // move absolute item X to virtual item X
    i := self.qStart;
    k := 0;
    repeat
        s1 := self.GetFilenameAbsolute(i) + BAK_EXT;
        s2 := self.GetFilenameAbsolute(k);
        if FileExists(s1) then
            RenameFile(s1, s2);

        s1 := self.GetFilenameClipAbsolute(i) + BAK_EXT;
        s2 := self.GetFilenameClipAbsolute(k);
        if FileExists(s1) then
            RenameFile(s1, s2);



        s1 := self.GetFilenameClipAbsolute(i);
        s1 := leftstr(s1,length(s1)-4) + FILE_EXT + BAK_EXT;
        s2 := self.GetFilenameClipAbsolute(k);
        if FileExists(s1) then
            RenameFile(s1, s2);



        s1 := self.GetFilenameIcon1Absolute(i) + BAK_EXT;
        s2 := self.GetFilenameIcon1Absolute(k);
        if FileExists(s1) then
            RenameFile(s1, s2);

        s1 := self.GetFilenameIcon2Absolute(i) + BAK_EXT;
        s2 := self.GetFilenameIcon2Absolute(k);
        if FileExists(s1) then
            RenameFile(s1, s2);

        i := Cardinal(i + 1) mod self.qSize;
        inc(k);
    until (Cardinal(i) = self.qStart);
    self.qStart := 0;
end;
}


{
function TPagedStringQueue.GetFilenameOld(index: cardinal): string;
begin
    index := self.IndexTranslate(index);
    result := self.GetFilenameAbsolute(index);
end;
function TPagedStringQueue.GetFilenameNew(index: cardinal): string;
begin
    index := self.IndexTranslate(index);
    result := self.GetFilenameAbsolute(index);
end;
}



function TPagedStringQueue.GetFilenameClip(index: cardinal): string;
begin
    index := self.IndexTranslate(index);
    result := self.GetFilenameClipAbsolute(index);
end;



function TPagedStringQueue.GetFilenameAbsoluteOld(
  absoluteIndex: cardinal): string;
begin
    result := self.cache + IntToHex(absoluteIndex,8) + PAGED_EXT;
end;

function TPagedStringQueue.GetFilenameAbsolute(absoluteindex: cardinal): string;
begin
    result := self.cache + IntToHex(absoluteindex,8) + FILE_EXT;
end;




function TPagedStringQueue.GetFilenameClipAbsolute(
  absoluteIndex: cardinal): string;
begin
    result := ci.GetFilename(self.cache, absoluteIndex);
end;
function TPagedStringQueue.GetFilenameIcon1Absolute(
  absoluteIndex: cardinal): string;
begin
    result := ci.GeticonFilename(self.iconcache, absoluteIndex, '-m');
end;
function TPagedStringQueue.GetFilenameIcon2Absolute(
  absoluteIndex: cardinal): string;
begin
    result := ci.GeticonFilename(self.iconcache, absoluteIndex, '-c');
end;

function TPagedStringQueue.GetItemAbsolute(
  absoluteIndex: cardinal): string;
var f : file;
    filename, newfilename, s : string;
    i : integer;
begin
    // default to the new format
    // convert old format to new format automatically

    filename := self.GetFilenameAbsoluteOld(absoluteIndex);
    newfilename := self.GetFilenameAbsolute(absoluteIndex);
    if fileexists(newfilename) then begin
        filename := newfilename;
    end else begin
        if fileexists(filename) then begin
            renamefile(filename, newfilename);
            filename := newfilename;
        end;
    end;

    //
    // read the item text from the cache
    //
    result := '';
    s := '';
    if (FileExists(filename)) then begin
        try
            Assign(f, filename);
            FileMode := 0;
            Reset(f, 1);
            i := filesize(f);
            if i <> 0 then begin
                setlength(s,i);
                blockread(f, s[1], i);
            end;
            result := s;
            closefile(f);
        except
            on e : exception do begin
                UnitMisc.AppendLog('GetItemAbsolute: ' + e.Message );
            end;
        end;
    end;
end;
function TPagedStringQueue.getItemClipAbsolute(
  absoluteIndex: cardinal): TClipItem;
begin
    result := nil;

    if FileExists(self.GetFilenameClipAbsolute(absoluteIndex)) then begin
        result := TClipItem.Create;
        result.LoadFromFIle(self.cache, absoluteIndex);
    end;
end;







{ TTempClipQueue }

procedure TTempClipQueue.AddNoSizeCheck(s: string; ci: TClipItem);
begin
    sl.AddObject(s, ci);
end;

procedure TTempClipQueue.InsertAtStart(ci: TClipItem);
begin
    sl.InsertObject(0, ci.GetAsText, ci);
    self.SetQueueSize(self.qSize);
end;

procedure TTempClipQueue.DeleteItem(index: cardinal);
begin
    self.DeleteItem2(index, false);
end;

procedure TTempClipQueue.DeleteItem2(index: cardinal; NoFree : boolean = false);
var ci : TClipItem;
begin
     ci := TClipITem(sl.Objects[index]);

    // Add the text to the removed items
    // ClipItem also has global memory to free
    //RemovedQueue.InsertAtStart(self.GetItemText(index), ci);

    if (ci <> nil) then begin
        UnitMisc.AppendLog('^Deleting Icon^');
        if not NoFree then MyDestroyIcon(ci.CData.GetHICONAbsolute );
    end;
    if not NoFree then
        MyFree(ci);

    // this MUST be the ONLY place an item is deleted
    sl.Delete(index);
end;

procedure ResyncTempClipQueue;
var i : integer;
begin
    TempClipQueue.ClearQueue;
    
    if removedqueue.GetQueueCount <> 0 then begin
        if TempClipQueue.GetQueueSize <> 0 then begin
            for i := (RemovedQUeue.GetQueueCount - 1) downto
                (max(RemovedQueue.GetQueueCount - cardinal(FrmConfig.udRecentlyRemoved.Position), 0)) do begin
                TempClipQueue.AddNoSizeCheck(
                    RemovedQueue.GetItemText( i ),
                    RemovedQueue.GetItemClip( i )
                );
            end;
        end;
  end;
end;




{////////////////////}
{//}initialization{//}
{////////////////////}
begin

    RemovedQueue := TPagedStringQueue.Create('removed.txt', 'removedcache\');
    ClipQueue := TClipQueue.Create;

    TempClipQueue := TTempClipQueue.Create;
    TempClipQueue.SetQueueSize(5);



    ClipDataDefaultIcon := LoadIcon(0, IDI_APPLICATION);
end;
{//////////////////}
{//}finalization{//}
{//////////////////}
begin
    RemovedQueue.Free;
    ClipQueue.Free;
end;
end.
